"""
Test the functions related to the internal interface implementation and the 'Interface' interface itself
"""

import logging

from cppython_core.schema import (
    Generator,
    GeneratorConfiguration,
    GeneratorData,
    PyProject,
)
from pytest_mock import MockerFixture

from cppython.data import default_pyproject
from cppython.project import Project, ProjectBuilder, ProjectConfiguration


class MockGeneratorData(GeneratorData):
    """
    TODO
    """

    check: bool


class TestProject:
    """
    TODO
    """

    def test_construction(self, mocker: MockerFixture):
        """
        Makes sure the project can be created from the default PyProject data
        """

        interface_mock = mocker.MagicMock()
        configuration = ProjectConfiguration()
        Project(configuration, interface_mock, default_pyproject.dict(by_alias=True))


class TestBuilder:
    """
    TODO
    """

    def test_plugin_gather(self):
        """
        TODO
        """

        configuration = ProjectConfiguration()
        builder = ProjectBuilder(configuration)
        plugins = builder.gather_plugins(Generator)

        assert len(plugins) == 0

    def test_generator_data_construction(self, mocker: MockerFixture):
        """
        TODO
        """

        configuration = ProjectConfiguration()
        builder = ProjectBuilder(configuration)
        model_type = builder.generate_model([])

        assert model_type.__base__ == PyProject

        generator_type = mocker.Mock(spec=Generator)
        generator_type.name.return_value = "mock"
        generator_type.data_type.return_value = MockGeneratorData

        model_type = builder.generate_model([generator_type])

        project_data = default_pyproject.dict(by_alias=True)

        mock_data = MockGeneratorData(check=True)
        project_data["tool"]["cppython"]["mock"] = mock_data.dict(by_alias=True)
        result = model_type(**project_data)

        assert result.tool is not None
        assert result.tool.cppython is not None

        assert result.tool.cppython.mock.check

    def test_generator_creation(self, mocker: MockerFixture):
        """
        TODO
        """

        configuration = ProjectConfiguration()
        builder = ProjectBuilder(configuration)

        generator_configuration = GeneratorConfiguration(logging.getLogger(__name__))
        generators = builder.create_generators([], generator_configuration, default_pyproject)

        assert not generators

        generator = mocker.Mock(spec=Generator)
        generators = builder.create_generators([generator], generator_configuration, default_pyproject)

        assert len(generators) == 1



"""
Test the functions related to the internal interface implementation and the 'Interface' interface itself
"""

import pytest
from click.testing import CliRunner
from cppython_core.schema import API
from pytest_cppython.plugin import InterfaceUnitTests
from pytest_mock.plugin import MockerFixture

from cppython.console import Config, ConsoleInterface, cli
from cppython.data import default_pyproject


class TestCLIInterface(InterfaceUnitTests):
    """
    The tests for our CLI interface
    """

    @pytest.fixture(name="interface")
    def fixture_interface(self) -> ConsoleInterface:
        """
        Override of the plugin provided interface fixture.

        Returns:
            ConsoleInterface -- The Interface object to use for the CPPython defined tests
        """
        return ConsoleInterface()

    # Grab the API methods and parameterize them for automatic testing of the entry_points
    method_list = [func for func in dir(API) if callable(getattr(API, func)) and not func.startswith("__")]

    @pytest.mark.parametrize("command", method_list)
    def test_command(self, command: str, mocker: MockerFixture):
        """
        _summary_

        Arguments:
            command {str} -- The CLI command with the same name as the CPPython API call
            mocker {MockerFixture} -- pytest-mock fixture
        """
        # Patch the project initialization
        mocker.patch("cppython.project.Project.__init__", return_value=None)

        # Patch the reading of data
        mocker.patch("cppython.console._create_pyproject", return_value=default_pyproject)

        config = Config()

        # Patch out the implementation
        mocked_command = mocker.patch(f"cppython.project.Project.{command}")

        runner = CliRunner()
        result = runner.invoke(cli, [command], obj=config, catch_exceptions=False)

        assert result.exit_code == 0
        mocked_command.assert_called_once()

    def test_config(self):
        """
        TODO
        """

        Config()

    def test_verbosity(self):
        """
        TODO
        """

        config = Config()

        runner = CliRunner()
        result = runner.invoke(cli, "-v info", obj=config, catch_exceptions=False)

        assert result.exit_code == 0

        assert config.configuration.verbosity



"""
A click CLI for CPPython interfacing
"""

from logging import Logger
from pathlib import Path
from typing import Any, Type

import click
import tomlkit
from cppython_core.schema import GeneratorDataType, Interface

from cppython.project import Project, ProjectConfiguration


def _create_pyproject() -> dict[str, Any]:

    # Search for a path upward
    path = Path.cwd()

    while not path.glob("pyproject.toml"):
        if path.is_absolute():
            assert (
                False
            ), "This is not a valid project. No pyproject.toml found in the current directory or any of its parents."

    path = Path(path / "pyproject.toml")

    # Load file
    data = tomlkit.loads(path.read_text(encoding="utf-8"))

    # Interpret and validate data
    return data


class Config:
    """
    The data object that will be expanded alongside 'pass_obj'
    """

    def __init__(self):
        self.pyproject_data = _create_pyproject()
        self.interface = ConsoleInterface()
        self.configuration = ProjectConfiguration()

    def create_project(self) -> Project:
        """
        TODO
        """
        return Project(self.configuration, self.interface, self.pyproject_data)


pass_config = click.make_pass_decorator(Config, ensure=True)


@click.group()
@click.option("-v", "--verbose", count=True, help="Print additional output")
@pass_config
def cli(config, verbose: int):
    """
    entry_point group for the CLI commands
    """
    config.configuration.verbosity = verbose


@cli.command()
@pass_config
def info(config):
    """
    TODO
    """
    config.create_project()


@cli.command()
@pass_config
def install(config):
    """
    TODO
    """
    project = config.create_project()
    project.install()


@cli.command()
@pass_config
def update(config):
    """
    TODO
    """
    project = config.create_project()
    project.update()


@cli.command()
@pass_config
def build(config):
    """
    TODO
    """
    project = config.create_project()
    project.build()


class ConsoleInterface(Interface):
    """
    Interface implementation to pass to the project
    """

    def read_generator_data(self, generator_data_type: Type[GeneratorDataType]) -> GeneratorDataType:
        """
        Requests generator information
        """
        return generator_data_type()

    def write_pyproject(self) -> None:
        """
        Write output
        """

    def register_logger(self, logger: Logger) -> None:
        """
        TODO
        """



"""
The central delegation of the CPPython project
"""

import logging
from dataclasses import dataclass
from importlib import metadata
from typing import Any, Type, TypeVar

from cppython_core.schema import (
    API,
    CPPythonData,
    Generator,
    GeneratorConfiguration,
    Interface,
    Plugin,
    PyProject,
    ToolData,
)
from pydantic import create_model


@dataclass
class ProjectConfiguration:
    """
    TODO
    """

    _verbosity: int = 0

    @property
    def verbosity(self) -> int:
        """
        TODO
        """
        return self._verbosity

    @verbosity.setter
    def verbosity(self, value: int) -> None:
        """
        TODO
        """
        self._verbosity = min(max(value, 0), 2)


class ProjectBuilder:
    """
    TODO
    """

    def __init__(self, configuration: ProjectConfiguration) -> None:
        self.configuration = configuration

    DerivedPlugin = TypeVar("DerivedPlugin", bound=Plugin)

    def gather_plugins(self, plugin_type: Type[DerivedPlugin]) -> list[Type[DerivedPlugin]]:
        """
        TODO
        """
        plugins = []
        entry_points = metadata.entry_points(group=f"cppython.{plugin_type.plugin_group()}")

        for entry_point in entry_points:
            loaded_plugin_type = entry_point.load()
            if issubclass(loaded_plugin_type, plugin_type) & (loaded_plugin_type is not plugin_type):
                plugins.append(loaded_plugin_type)

        return plugins

    def generate_model(self, plugins: list[Type[Generator]]) -> Type[PyProject]:
        """
        TODO: Proper return type hint
        """
        plugin_fields = {}
        for plugin_type in plugins:
            plugin_fields[plugin_type.name()] = (plugin_type.data_type(), ...)

        extended_cppython_type = create_model(
            "ExtendedCPPythonData",
            **plugin_fields,
            __base__=CPPythonData,
        )

        extended_tool_type = create_model(
            "ExtendedToolData",
            cppython=(extended_cppython_type, ...),
            __base__=ToolData,
        )

        return create_model(
            "ExtendedPyProject",
            tool=(extended_tool_type, ...),
            __base__=PyProject,
        )

    def create_generators(
        self, plugins: list[Type[Generator]], configuration: GeneratorConfiguration, pyproject: PyProject
    ) -> list[Generator]:
        """
        TODO
        """
        _generators = []
        for plugin_type in plugins:
            _generators.append(plugin_type(configuration, pyproject))

        return _generators


class Project(API):
    """
    The object constructed at each entry_point
    """

    def __init__(
        self, configuration: ProjectConfiguration, interface: Interface, pyproject_data: dict[str, Any]
    ) -> None:

        self._enabled = False
        self._configuration = configuration
        self._pyproject = None

        levels = [logging.WARNING, logging.INFO, logging.DEBUG]

        self._logger = logging.getLogger("cppython")
        self._logger.setLevel(levels[configuration.verbosity])

        interface.register_logger(self._logger)

        self._logger.info("Initializing project")

        builder = ProjectBuilder(self.configuration)
        plugins = builder.gather_plugins(Generator)

        if not plugins:
            self._logger.info("No generator plugin was found")
            return

        extended_pyproject_type = builder.generate_model(plugins)
        self._pyproject = extended_pyproject_type(**pyproject_data)

        if self.pyproject is None:
            self._logger.info("Data is not defined")
            return

        if self.pyproject.tool is None:
            self._logger.info("Table [tool] is not defined")
            return

        if self.pyproject.tool.cppython is None:
            self._logger.info("Table [tool.cppython] is not defined")
            return

        self._enabled = True

        self._interface = interface

        generator_configuration = GeneratorConfiguration(self._logger)
        self._generators = builder.create_generators(plugins, generator_configuration, self.pyproject)

        self._logger.info("Initialized project")

    @property
    def enabled(self) -> bool:
        """
        TODO
        """
        return self._enabled

    @property
    def configuration(self) -> ProjectConfiguration:
        """
        TODO
        """
        return self._configuration

    @property
    def pyproject(self) -> PyProject | None:
        """
        TODO
        """
        return self._pyproject

    def download(self):
        """
        Download the generator tooling if required
        """
        if self._enabled:
            base_path = self.pyproject.tool.cppython.install_path

            for generator in self._generators:

                path = base_path / generator.name()

                path.mkdir(parents=True, exist_ok=True)

                if not generator.generator_downloaded(path):
                    self._logger.info(f"Downloading the {generator.name()} tool")

                    # TODO: Make async with progress bar
                    generator.download_generator(path)
                    self._logger.info("Download complete")

    # API Contract

    def install(self) -> None:
        if self._enabled:
            self._logger.info("Installing project")
            self.download()

            for generator in self._generators:
                generator.install()

    def update(self) -> None:
        if self._enabled:
            self._logger.info("Updating project")

            for generator in self._generators:
                generator.update()

    def build(self) -> None:
        if self._enabled:
            self._logger.info("Building project")

            for generator in self._generators:
                generator.build()



"""
Defaulted data to help testing
"""

from cppython_core.schema import PEP621, CPPythonData, PyProject, TargetEnum, ToolData

default_pep621 = PEP621(name="test_name", version="1.0")

# CMake is a default plugin
default_cppython_data = CPPythonData(**{"target": TargetEnum.EXE})
default_tool_data = ToolData(**{"cppython": default_cppython_data})
default_pyproject = PyProject(**{"project": default_pep621, "tool": default_tool_data})



"""
Test the integrations related to the internal interface implementation and the 'Interface' interface itself
"""

import pytest
from pytest_cppython.plugin import InterfaceIntegrationTests

from cppython.console import ConsoleInterface


class TestCLIInterface(InterfaceIntegrationTests):
    """
    The tests for our CLI interface
    """

    @pytest.fixture(name="interface")
    def fixture_interface(self):
        """
        Override of the plugin provided interface fixture.

        Returns:
            ConsoleInterface -- The Interface object to use for the CPPython defined tests
        """
        return ConsoleInterface()
